#include "AlarmSignal.h"
#include <eventconsumer.h>
#include "AlarmNotify.h"
#include "AlarmData.h"
#include "PushThread.h"
#include "SubscribeThread.h"


namespace AlarmNotify_ns
{

    std::mutex AlarmSignal::static_mutex;

    //=============================================================================
    //=============================================================================

    ArchiveCB::ArchiveCB(Tango::DeviceImpl *dev):Tango::LogAdapter(dev)
    {
        _device=dev;
    }

//=============================================================================
/**
 *	Attribute and Event management
    */
//=============================================================================
void ArchiveCB::push_event(Tango::EventData *data)
{
    AlarmNotify *_dev = static_cast<AlarmNotify *>(_device);
	auto now = std::chrono::system_clock::now();
    auto min_warmup = min(90, 1+2*_dev->subscribePeriod);//TODO: configurable
	if(std::chrono::duration_cast<std::chrono::seconds>(now - _dev->startup) < std::chrono::seconds(min_warmup))
	{
		INFO_STREAM<< __func__ << ": skipping notification for " << data->attr_name <<" during WARM-UP time elapsed="<< std::chrono::duration_cast<std::chrono::seconds>(now - _dev->startup).count() << "<" << min_warmup << endl;
		return;
	}
	else
	{
		//INFO_STREAM<< __func__ << ": WARMED-UP notification for " << data->attr_name <<" , elapsed="<< std::chrono::duration_cast<std::chrono::seconds>(now - _dev->startup).count() << endl;
	}
    //time_t	t = time(NULL);
    //DEBUG_STREAM << __func__<<": Event '"<<data->attr_name<<"' id="<<omni_thread::self()->id() << "  Received at " << ctime(&t);
    AlarmStringUtils::fix_tango_host(data->attr_name);	//TODO: why sometimes event arrive without fqdn ??
    string handler, tag;
    AlarmStringUtils::split_tag_handler(data->attr_name, handler, tag);
    try
    {
        auto signal = _dev->shared->get_signal(data->attr_name);

        AlarmEventDataType ev_data_type;
        ev_data_type.attr_name = data->attr_name;
        try
        {
            auto conf = signal->get_signal_config();
        
            ev_data_type.data_type = conf.data_type;
            ev_data_type.data_format = conf.data_format;
            ev_data_type.write_type = conf.write_type;
            ev_data_type.max_dim_x = conf.max_dim_x;
            ev_data_type.max_dim_y = conf.max_dim_y;
        
        }
        catch (Tango::DevFailed &e)
        {
            INFO_STREAM<< __func__ << ": FIRST exception in get_config: " << data->attr_name <<" ev_data_type.data_type="<<ev_data_type.data_type<<" err="<<e.errors[0].desc<< endl;
            return;
        }

        //	Check if event is an error event.
        if(data->err)
        {
            std::string error(data->errors[0].desc);
            signal->set_error(error);

            INFO_STREAM<< __func__ << ": Exception on " << data->attr_name << endl;
            INFO_STREAM << data->errors[0].desc  << endl;
            try
            {
                signal->set_nok_event();
            }
            catch(Tango::DevFailed &e)
            {
                WARN_STREAM << __func__ << " Unable to set_nok_event: " << e.errors[0].desc << "'"<<endl;
            }

            try
            {
                if(!(signal->is_running() && signal->is_first_err()))
                {
                    return;
                }
            }
            catch(Tango::DevFailed &e)
            {
                WARN_STREAM << __func__ << " Unable to check if is_running: " << e.errors[0].desc << "'"<<endl;
            }
            try
            {
                signal->set_first_err();
            }
            catch(Tango::DevFailed &e)
            {
                WARN_STREAM << __func__ << " Unable to set first err: " << e.errors[0].desc << "'"<<endl;
            }
        }
        else
        {
            try
            {
                signal->set_ok_event();	//also reset first_err
            }
            catch(Tango::DevFailed &e)
            {
                WARN_STREAM << __func__ << " Unable to set_ok_event: " << e.errors[0].desc << "'"<<endl;
            }
            //	Check if already OK
            if (signal->get_state() != Tango::ON)
            {
                signal->set_on();
            }

            //if attribute stopped, just return
            try
            {
                if(!signal->is_running() && !signal->is_first())
                {
                    return;
                }
            }
            catch(Tango::DevFailed &e)
            {
                WARN_STREAM << __func__ << " Unable to check if is_running: " << e.errors[0].desc << "'"<<endl;
            }
            try
            {
                if(signal->is_first())
                    signal->set_first();
            }
            catch(Tango::DevFailed &e)
            {
                WARN_STREAM << __func__ << " Unable to set first: " << e.errors[0].desc << "'"<<endl;
            }
        }

        //OK only with C++11:
        //Tango::EventData	*cmd = new Tango::EventData(*data);
        //OK with C++98 and C++11:
        auto *dev_attr_copy = new Tango::DeviceAttribute();
        if (!data->err)
        {
            dev_attr_copy->deep_copy(*(data->attr_value));
        }
        const auto &hdlr = _dev->handlerMap.find(handler);
        unordered_map<string,string> alarm_info;
        Tango::DevVarStringArray arg;
        arg.length(1);
        if(hdlr != _dev->handlerMap.end())
        {
            Tango::DeviceData Din,Dout;
            try
            {
                arg[0] = CORBA::string_dup(tag.c_str());
                Din << arg;
                Dout = hdlr->second.dp->command_inout("GetAlarmInfo",Din);
                vector<string> tmp;
                Dout >> tmp;
                AlarmStringUtils::string_vector2map(tmp, "=", alarm_info);
            }
            catch(Tango::DevFailed &ee)
            {
                INFO_STREAM << __func__<<": Event '"<<data->attr_name<<"' Cannot read GetAlarmInfo=" << ee.errors[0].desc;
            }
        }
        auto *ev_data = new Tango::EventData(data->device,data->attr_name, data->event, dev_attr_copy, data->errors);

        auto cmd = make_unique<AlarmData>(*signal, ev_data, ev_data_type, alarm_info);
        _dev->push_thread->push_back_cmd(std::move(cmd));
    }
    catch(Tango::DevFailed &e)
    {
        ERROR_STREAM << __func__<<": Event '"<<data->attr_name<<"' NOT FOUND in signal list, err=" << e.errors[0].desc << endl;
        return;
    }
}



    AlarmSignal::AlarmSignal(Tango::DeviceImpl *dev
            , const std::string& signame) : LogAdapter(dev)
                                                       , name(signame)
                                                       , config_set(false)
    {
        // on name, split device name and attrib name
        string::size_type idx = name.find_last_of('/');
        if (idx == string::npos)
        {
            Tango::Except::throw_exception(
                    (const char *)"SyntaxError",
                    "Syntax error in signal name " + name,
                    (const char *)"SharedData::add()");
        }
        // Build Hdb Signal object
        devname = name.substr(0, idx);
        attname = name.substr(idx+1);
        status = "NOT connected";
        attr = nullptr;
        running = false;
        stopped = true;

        using std::placeholders::_1;
        using std::placeholders::_2;
        using std::placeholders::_3;
    }

    void AlarmSignal::remove_callback()
    {
        unsubscribe_event(event_id);

        WriterLock lock(siglock);
        event_id = ERR;
        archive_cb.reset(nullptr);
        attr.reset(nullptr);
    }

    auto AlarmSignal::subscribe_events(Tango::DeviceImpl *_dev) -> void
    {
        AlarmNotify *dev = static_cast<AlarmNotify *>(_dev);
        if(attr)
        {
            WriterLock lock(siglock);
            archive_cb = make_unique<ArchiveCB>(dev);

            // Subscribe att_event
            try
            {
                try
                {
                    event_id = attr->subscribe_event(
                            Tango::ARCHIVE_EVENT,
                            archive_cb.get(),
                            /*stateless=*/false);
                }
                catch (Tango::DevFailed &e)
                {
                    Tango::Except::print_exception(e);
                    if (dev->subscribe_change)
                    {
                        INFO_STREAM <<__func__<< " " << name << "->subscribe_event EXCEPTION, try CHANGE_EVENT" << endl;
                        event_id = attr->subscribe_event(
                                Tango::CHANGE_EVENT,
                                archive_cb.get(),
                                /*stateless=*/false);
                        INFO_STREAM <<__func__<< " " << name << "->subscribe_event CHANGE_EVENT SUBSCRIBED" << endl;
                    } 
                    else 
                    {
                        throw(e);
                    }
                }

                DEBUG_STREAM << name << "(id="<< event_id <<"): Subscribed Event" << endl;
            }
            catch(Tango::DevFailed& e)
            {
                INFO_STREAM <<__func__<< " attr->subscribe_event EXCEPTION:" << endl;
                Tango::Except::print_exception(e);
                status = e.errors[0].desc;
                event_id = ERR;
            }

        }
    }

    auto AlarmSignal::init() -> void
    {
        WriterLock lock(siglock);
        event_id = ERR;
        evstate = Tango::ALARM;
        ok_events.reset();
        nok_events.reset();
        first_err = true;
    }

    auto AlarmSignal::start() -> void
    {
        WriterLock lock(siglock);
        status = "NOT connected";
        //DEBUG_STREAM << "created proxy to " << signame << endl;
        //	create Attribute proxy
        if(!attr)
        {
            attr = make_unique<Tango::AttributeProxy>(name);	//TODO: OK out of siglock? accessed only inside the same thread?
            DEBUG_STREAM << "SharedData::"<<__func__<<": signame="<<name<<" created proxy"<< endl;
        }
        
        try
        {
            get_signal_config();
        }
        catch (Tango::DevFailed &e)
        {
            status = e.errors[0].desc;
            INFO_STREAM <<"AlarmSignal::"<<__func__<< " ERROR for " << name << " in get_config err=" << e.errors[0].desc << endl;
        }        
    }

    auto AlarmSignal::get_error() -> std::string
    {
        ReaderLock lock(siglock);

        if (evstate != Tango::ON && is_running())
        {
            return status;
        }
        ReaderLock lk(dblock);
        if(dbstate == Tango::ALARM)
        {
            return dberror;
        }

        return "";
    }

    auto AlarmSignal::set_ok_event() -> void
    {
        ok_events.increment();
        
        {
        WriterLock lock(siglock);

        evstate = Tango::ON;
        status = "Event received";
        first_err = true;
        }
    }

    auto AlarmSignal::set_nok_event() -> void
    {
        nok_events.increment();
    }

    auto AlarmSignal::set_nok_notification(const std::string& error) -> void
    {
        noknotif_events.increment();
        {
        WriterLock lock(dblock);
        dbstate = Tango::ALARM;
        //dberror = STATUS_DB_ERROR;
        if(error.length() > 0)
            dberror += ": " + error;
        }
    }

    auto AlarmSignal::set_ok_notification() -> void
    {
        {
            WriterLock lock(dblock);

            dbstate = Tango::ON;
            dberror = "";
        }
        oknotif_events.increment();
    }

    auto AlarmSignal::set_sent_notification() -> void
    {
        {
            WriterLock lock(dblock);
        }
        sentnotif_events.increment();
    }

    auto AlarmSignal::set_alarm_state(const std::string &_alarm_state) -> void
    {
        WriterLock lock(dblock);
        alarm_state = _alarm_state;       
    }

    auto AlarmSignal::get_signal_config() -> AlarmSignal::SignalConfig
    {
        if(!config_set)
        {
            WriterLock lock(siglock);
            if(attr)
            {
                Tango::AttributeInfo info = attr->get_config();

                config.data_type = info.data_type;
                config.data_format = info.data_format;
                config.write_type = info.writable;
                config.max_dim_x = info.max_dim_x;
                config.max_dim_y = info.max_dim_y;

                config_set = true;
            }
            else
            {
                Tango::Except::throw_exception(
                        (const char *)"AttributeNotStarted",
                        "Attribute " + name + " not started.",
                        (const char *)__func__);
            }
        }
        ReaderLock lock(siglock);
        return config;
    }

    void AlarmSignal::unsubscribe_event(const int event_id)
    {
        ReaderLock lock(siglock);
        if(event_id != ERR && attr != nullptr)
        {
            //unlocking, locked in SharedData::stop but possible deadlock if unsubscribing remote attribute with a faulty event connection
            //siglock.writerOut();
            attr->unsubscribe_event(event_id);
            //siglock.writerIn();
        }
    }
}
