#ifndef ALARMSIGNAL_H
#define ALARMSIGNAL_H

#include <deque>
#include <chrono>
#include <string>
#include <memory>
#include <vector>
#include <thread>
#include <mutex>
#include <condition_variable>
#include <atomic>
#include <tango.h>


namespace AlarmNotify_ns
{

template<typename T, typename... Args>
std::unique_ptr<T> make_unique(Args&&... args)
{
    return std::unique_ptr<T>(new T(std::forward<Args>(args)...));
}

//==========================================================
/**
 * Class Description:
 * This class manage archive events
 */
//==========================================================
class ArchiveCB : public Tango::CallBack, public Tango::LogAdapter
{
public:
        Tango::DeviceImpl *_device;
//      HdbSignal       *signal;
        ArchiveCB(Tango::DeviceImpl *dev);// { hdb_dev=dev; };

        virtual void push_event(Tango::EventData *data);
};


    class AlarmSignal : public Tango::LogAdapter
    {
        static constexpr int ERR = -1;

        public:
        
        std::string name;

        explicit AlarmSignal(Tango::DeviceImpl *dev, const std::string& name);

        struct SignalConfig
        {
            int write_type;
            int max_dim_x;
            int max_dim_y;
            int data_type;
            Tango::AttrDataFormat data_format;
        };

        void remove_callback();

        void subscribe_events(Tango::DeviceImpl *dev);      

        auto init() -> void;

        auto start() -> void;

        auto is_running() -> bool
        {
            ReaderLock lock(siglock);
            return running; 
        }

        auto is_not_subscribed() -> bool
        {
            ReaderLock lock(siglock);
            return event_id == ERR && !is_stopped();
        }

        auto is_on_error() -> bool
        {
            ReaderLock lock(siglock);
            ReaderLock lk(dblock);
            return (evstate == Tango::ALARM && is_running()) || dbstate == Tango::ALARM;
        }

        auto is_on() -> bool
        {
            ReaderLock lock(siglock);
            return evstate == Tango::ON && is_running();
        }

        auto is_not_on_error() -> bool
        {
            ReaderLock lock(siglock);
            ReaderLock lk(dblock);
            return (evstate == Tango::ON || (evstate == Tango::ALARM && !is_running())) && dbstate != Tango::ALARM;
        }

        auto get_error() -> std::string;

        auto get_ok_events() -> unsigned int
        {
            return get_events(ok_events);
        }

        auto get_nok_events() -> unsigned int
        {
            return get_events(nok_events);
        }
        
        auto get_nok_notification() -> unsigned int
        {
            return get_events(noknotif_events);
        }
        
        auto get_ok_notification() -> unsigned int
        {
            return get_events(oknotif_events);
        }

        auto get_sent_notification() -> unsigned int
        {
            return get_events(sentnotif_events);
        }

        auto get_total_events() -> unsigned int
        {
            return get_ok_events() + get_nok_events();
        }

        auto get_last_ok_event() -> std::chrono::time_point<std::chrono::system_clock>
        {
            return get_last_event(ok_events);
        }

        auto get_last_nok_event() -> std::chrono::time_point<std::chrono::system_clock>
        {
            return get_last_event(nok_events);
        }

        auto get_last_ok_notification() -> std::chrono::time_point<std::chrono::system_clock>
        {
            return get_last_event(oknotif_events);
        }

        auto get_last_nok_notification() -> std::chrono::time_point<std::chrono::system_clock>
        {
            return get_last_event(noknotif_events);
        }

        auto get_rate_ok_notification(Tango::DevLong period) -> size_t
        {
            return get_rate_event(oknotif_events, period);
        }

        auto get_last_sent_notification() -> std::chrono::time_point<std::chrono::system_clock>
        {
            return get_last_event(sentnotif_events);
        }

        auto set_ok_event() -> void;

        auto set_nok_event() -> void;

        auto set_nok_periodic_event() -> void;

        auto set_nok_notification(const std::string& error) -> void;
        
        auto set_ok_notification() -> void;

        auto set_sent_notification() -> void;

        auto set_alarm_state(const std::string &_alarm_state) -> void;

        auto get_status() -> std::string
        {
            ReaderLock lock(siglock);
            ReaderLock lk(dblock);
            std::stringstream ret;
            ret << status;
            if(dbstate == Tango::ALARM)
            {
                ret << std::endl << dberror;
            }
            return ret.str();
        }

        auto get_alarm_state() -> std::string
        {
            ReaderLock lock(siglock);
            return alarm_state;
        }

        auto get_state() -> Tango::DevState
        {
            ReaderLock lock(siglock);
            ReaderLock lk(dblock);
            if(dbstate == Tango::ALARM)
                return Tango::ALARM;
            return evstate;
        }

        auto reset_statistics() -> void
        {
            ok_events.reset();
            nok_events.reset();
            noknotif_events.reset();
            oknotif_events.reset();
            sentnotif_events.reset();
        }

        auto get_signal_config() -> SignalConfig;

        auto is_stopped() -> bool
        {
            ReaderLock lock(siglock);
            return stopped; 
        }

        auto set_running() -> void
        {
            {
                WriterLock lock(siglock);
                running = true;
                stopped = false;
            }
        }

        auto set_stopped() -> void
        {
            {
                WriterLock lock(siglock);
                running = false;
                stopped = true;
            }
        }

        auto is_first() -> bool
        {
            ReaderLock lock(siglock);
            return first;
        }

        auto is_first_err() -> bool
        {
            ReaderLock lock(siglock);
            return first_err;
        }

        auto set_first() -> void
        {
            WriterLock lock(siglock);
            first = false;
        }

        auto set_first_err() -> void
        {
            WriterLock lock(siglock);
            first_err = false;
        }

        auto set_error(const std::string& err) -> void
        {
            {
                WriterLock lock(siglock);
                evstate = Tango::ALARM;
                status = err;
            }
        }

        auto set_on() -> void
        {
            {
                WriterLock lock(siglock);
                evstate = Tango::ON;
                status = "Subscribed";
            }
        }

        private:

        class EventCounter
        {
            friend class AlarmSignal;

            std::atomic_uint counter;
            //std::deque<std::chrono::time_point<std::chrono::system_clock>> timestamps;
            std::deque<timeval> timestamps;
            mutable std::mutex timestamps_mutex;

            EventCounter():counter(0)
            {
            }

            void increment()
            {
                ++counter;
                //timestamps.push_back(std::chrono::system_clock::now());
                timeval now;
                gettimeofday(&now, NULL);
                timestamps.push_back(now);
            }

            public:

            auto get_last_event() const -> std::chrono::time_point<std::chrono::system_clock>
            {
                std::lock_guard<mutex> lk(timestamps_mutex);
                if(timestamps.size()>0)
                {

                    timeval src = timestamps.back();

                    // again, trusting the value with only milliseconds accuracy
                    using dest_timepoint_type=std::chrono::time_point<
                        std::chrono::system_clock, std::chrono::milliseconds
                    >;
                    dest_timepoint_type converted{
                        std::chrono::milliseconds{
                        src.tv_sec*1000+src.tv_usec/1000
                        }
                    };

                    // this is to make sure the converted timepoint is indistinguishable by one
                    // issued by the system_clock
                    std::chrono::system_clock::time_point recovered =
                        std::chrono::time_point_cast<std::chrono::system_clock::duration>(converted)
                    ;

                    return recovered;
                }
                struct std::chrono::time_point<std::chrono::system_clock> ret;
                return ret;
            }

            auto get_rate_event(Tango::DevLong period) -> size_t
            {
                size_t ret=0;
                auto now = std::chrono::system_clock::now();
                std::lock_guard<mutex> lk(timestamps_mutex);
#if 0
                for(std::deque<std::chrono::time_point<std::chrono::system_clock>>::const_iterator t=timestamps.end()-1; t!=timestamps.begin(); t--)
                {
                    //auto seconds = std::chrono::duration_cast<std::chrono::seconds>(now-period);
                    std::cout << __func__ << ": looping time=" << std::chrono::system_clock::to_time_t(*t) << " now=" << std::chrono::system_clock::to_time_t(now) << endl;
                    if(*t > now-period)
                    {
                        ret++;
                        std::cout << __func__ << ": incremented ret="<<ret<<endl;
                    }
                    else
                    {
                        //A& ref = const_cast <A&>(a);
                        //std::deque<std::chrono::time_point<std::chrono::system_clock>>::iterator& tt = const_cast <std::deque<std::chrono::time_point<std::chrono::system_clock>>::iterator &>(t)
                        //timestamps.erase(t,timestamps.begin());
                        //timestamps.erase(t);
                        //while(t != timestamps.begin())
                        //    timestamps.erase(timestamps.begin());

/*std::deque<time_t> timestamps2;
std::deque<time_t>::const_iterator it;
                for(it=timestamps2.end()-1; it!=timestamps2.begin(); it--)
                {
                    timestamps2.erase(it);
                }
*/
                        break;
                    }
                }
#else

                for(std::deque<timeval>::const_iterator t=timestamps.end()-1; t!=timestamps.begin() && timestamps.size()>0; t--)
                {
                    //auto seconds = std::chrono::duration_cast<std::chrono::seconds>(now-period);
                    timeval now;
                    gettimeofday(&now, NULL);
                    //std::cout << __func__ << ": after gettimeofday now=" << now.tv_sec << endl;
                    //std::cout << __func__ << ": looping time=" << t->tv_sec << " now=" << now.tv_sec << endl;
                    if(t->tv_sec > now.tv_sec-period)
                    {
                        ret++;
                        //std::cout << __func__ << ": incremented ret="<<ret<<endl;
                    }
                    else
                    {
#if 0
						timeval tt;
						do
						{
							tt = timestamps.back();
							if(tt.tv_sec < now.tv_sec-period)
								timestamps.pop_back();
						} while (tt.tv_sec < now.tv_sec-period);
#endif						
						
						
                        //A& ref = const_cast <A&>(a);
                        //std::deque<std::chrono::time_point<std::chrono::system_clock>>::iterator& tt = const_cast <std::deque<std::chrono::time_point<std::chrono::system_clock>>::iterator &>(t)
                        //timestamps.erase(t,timestamps.begin());
						
                        //timestamps.erase(t);
                        while(t != timestamps.begin())
						{
							//std::cout << __func__ << ": before erasing, timestamps.size=" << timestamps.size() << endl;
                            timestamps.erase(timestamps.begin());
							//std::cout << __func__ << ": after erasing, timestamps.size=" << timestamps.size() << endl;
						}
						break;

/*std::deque<time_t> timestamps2;
std::deque<time_t>::const_iterator it;
                for(it=timestamps2.end()-1; it!=timestamps2.begin(); it--)
                {
                    timestamps2.erase(it);
                }
*/
                        //break;
                    }
                }
#endif

                return ret;
            }

            void reset()
            {
                counter = 0;
            };
        };
        static std::mutex static_mutex;

        ReadersWritersLock siglock;
        ReadersWritersLock dblock;
        
        std::string devname;
        std::string attname;
        SignalConfig config;
        bool config_set;
        std::string status;
        std::unique_ptr<Tango::AttributeProxy> attr;
        Tango::DevState evstate;
        bool first_err;
        bool first;
        std::unique_ptr<ArchiveCB> archive_cb;
        int event_id;
        EventCounter ok_events;
        EventCounter nok_events;
        
        // DB insertion stats and state
        EventCounter noknotif_events;
        EventCounter oknotif_events;
        EventCounter sentnotif_events;
        Tango::DevState dbstate;
        std::string dberror;
        std::string alarm_state;
        
        bool running;
        bool stopped;

        void unsubscribe_event(const int event_id);

        auto get_events(const EventCounter& c) -> unsigned int
        {
            return c.counter.load();
        }

        auto get_last_event(const EventCounter& c) -> std::chrono::time_point<std::chrono::system_clock>
        {
            return c.get_last_event();
        }
        auto get_rate_event(EventCounter& c, Tango::DevLong period) -> size_t
        {
            return c.get_rate_event(period);
        }

    };
}
#endif // HDBSIGNAL_H
