//=============================================================================
//
// file :        HdbEventHandler.h
//
// description : Include for the HDbDevice class.
//
// project :	Tango Device Server
//
// $Author: graziano $
//
//
//
//
// copyleft :    European Synchrotron Radiation Facility
//               BP 220, Grenoble 38043
//               FRANCE
//
//=============================================================================

#ifndef _SUBSCRIBE_THREAD_H
#define _SUBSCRIBE_THREAD_H

#include <tango.h>
#include <eventconsumer.h>
#include <stdint.h>
#include <time.h>
#include <string>
#include <deque>
#include <mutex>

/**
 * @author	$Author: graziano $
 * @version	$Revision: 1.5 $
 */

 //	constants definitions here.
 //-----------------------------------------------

namespace AlarmNotify_ns
{

class ArchiveCB;
class SubscribeThread;
class AlarmSignal;

//=========================================================
/**
 *	Shared data between DS and thread.
 */
//=========================================================
//class SharedData: public omni_mutex
class SharedData: public Tango::TangoMonitor, public Tango::LogAdapter
{
private:
	/**
	 *	HdbDevice object
	 */
	Tango::DeviceImpl *_device;

	bool	stop_it;
        bool initialized;
        omni_mutex init_mutex;
        omni_condition init_condition;

        auto is_same_signal_name(const std::string& name1, const std::string& name2) -> bool;

        vector<std::shared_ptr<AlarmSignal>>	signals;
	ReadersWritersLock      veclock;
	
        void add(std::shared_ptr<AlarmSignal> signal, bool start);
	auto add(const string &signame, bool start) -> std::shared_ptr<AlarmSignal>;
public:

	/**
	 * Constructor
	 */
	//SharedData(HdbDevice *dev):condition(this){ hdb_dev=dev; action=NOTHING; stop_it=false; initialized=false;};
	SharedData(Tango::DeviceImpl *dev);
	~SharedData();
	/**
	 * Add a new signal.
	 */
	void add(const string &signame);
	/**
	 * Remove a signal in the list.
	 */
	void remove(const string &signame, bool stop);
	/**
	 * Unsubscribe events for this signal
	 */
	void unsubscribe_events(std::shared_ptr<AlarmSignal>);
	/**
	 * Start saving on DB a signal.
	 */
	void start(const string &signame);
	/**
	 * Stop saving on DB a signal.
	 */
	void stop(const string &signame);
	/**
	 * Start saving on DB all signals.
	 */
	void start_all();
	/**
	 * Stop saving on DB all signals.
	 */
	void stop_all();
	/**
	 * Is a signal saved on DB?
	 */
	auto is_running(const string &signame) -> bool;
	/**
	 * Is a signal not subscribed?
	 */
	auto is_stopped(const string &signame) -> bool;
	/**
	 * Is a signal first event arrived?
	 */
	auto is_first(const string &signame) -> bool;
	/**
	 * Set a signal first event arrived
	 */
	void set_first(const string &signame);
	/**
	 * Is a signal first consecutive error event arrived?
	 */
	auto is_first_err(const string &signame) -> bool;
	/**
	 *	get signal by name.
	 */
	auto get_signal(const string &name) -> std::shared_ptr<AlarmSignal>;
	/**
	 * Subscribe achive event for each signal
	 */
	void subscribe_events();
	/**
	 * Unsubscribe achive event for each signal
         * and clear the signals list
	 */
	void clear_signals();
	/**
	 *	return number of signals to be subscribed
	 */
	auto nb_sig_to_subscribe() -> int;
	/**
	 *	Return the list of signals
	 */
	void get_sig_list(vector<string> &);
	/**
	 *	Return the list of errors
	 */
	auto get_error_list(vector<string> &) -> bool;
	/**
	 *	Return the list of event received
	 */
	void get_ev_counter_list(vector<uint32_t> &);
	/**
	 *	Return the complete, started and stopped lists of signals
	 */
	auto  get_lists(vector<string> &s_list, vector<string> &s_start_list, vector<string> &s_stop_list) -> bool;
	/**
	 *	Return the list of signals + status + freq
	 */
	void get_sig_list_status(vector<string> &);
	/**
	 *	Get the ok counter of event rx
	 */
	auto get_ok_event(const string &signame) -> uint32_t;
	/**
	 *	Get last okev timestamp
	 */
	auto get_last_okev(const string &signame) -> std::chrono::time_point<std::chrono::system_clock>;
	/**
	 *	Get the error counter of event rx
	 */
	auto get_nok_event(const string &signame) -> uint32_t;
	/**
	 *	Get last nokev timestamp
	 */
	auto get_last_nokev(const string &signame) -> std::chrono::time_point<std::chrono::system_clock>;
	/**
	 *	Get the error counter of db saving
		*/
	auto get_nok_notification(const string& signame) -> uint32_t;
	/**
	 *	Get last nokdb timestamp
		*/
	auto get_last_nokdb(const string& signame) -> std::chrono::time_point<std::chrono::system_clock>;
	/**
	 *	Reset statistic counters
	 */
	void reset_statistics();
	/**
	 *	Return ALARM if at list one signal is not subscribed.
	 */
	auto state() -> Tango::DevState;
	
	auto is_initialized() -> bool;
	auto get_if_stop() -> bool;
	void stop_thread();
	void wait_initialized();
};








//=========================================================
/**
 *	Create a thread retry to subscribe event.
 */
//=========================================================
class SubscribeThread: public omni_thread, public Tango::LogAdapter
{
private:
	/**
	 *	Shared data
	 */
	std::shared_ptr<SharedData> shared;
	int			period;
	/**
	 *	HdbDevice object
	 */
	Tango::DeviceImpl *_device;


public:
	SubscribeThread(Tango::DeviceImpl *dev);
	/**
	 *	Execute the thread loop.
	 *	This thread is awaken when a command has been received 
	 *	and falled asleep when no command has been received from a long time.
	 */
	void *run_undetached(void *);
	void start() {start_undetached();}
};


}	// namespace_ns

#endif	// _SUBSCRIBE_THREAD_H
